from os import chdir
from subprocess import PIPE, CompletedProcess, run

from ..const import INDENT, Emoji
from ..file_path import PypjFilePath
from ..resource import get_my_resource
from ..setting import PypjSetting
from .task import Task, TaskError


class Poetry(Task):
    def __init__(self, setting: PypjSetting, path: PypjFilePath) -> None:
        name = "Poetry"
        super().__init__(name, setting, path)
        self.setting = setting
        self.poetry_new_cmd = f"poetry new {setting.package_name}"
        if setting.use_src:
            self.poetry_new_cmd += " --src"
        self.venv_in_pj_cmd = f"poetry config virtualenvs.in-project {str(setting.venv_in_pj).lower()}"

    def execute(self) -> None:
        print(f"Task: Initialize package: {self.setting.package_name}")
        self.done_check()
        self.__new()
        self.__configure_poetry()
        chdir(self.path.package_dir)
        self.__add_dev_dependency()
        self.__add_dependency()
        self.__overwrite_files()
        print(f"{INDENT}Create : {self.setting.package_name} {Emoji.OK}")
        self.done()

    def __command(self, cmd: str) -> CompletedProcess:
        print(f"{INDENT}Command: {cmd} {Emoji.WAIT}", end="", flush=True)
        r = run(cmd, shell=True, stdout=PIPE, stderr=PIPE, text=True)
        status = Emoji.OK if r.returncode == 0 else Emoji.NG
        print(f"\r{INDENT}Command: {cmd} {status}", flush=True)
        return r

    def __new(self) -> None:
        # poetry new
        r = self.__command(self.poetry_new_cmd)
        if r.returncode != 0:
            raise TaskError(f"Failed to '{self.poetry_new_cmd}'")
        print(f"{INDENT}Poetry new done {Emoji.LETS_GO}")

    def __configure_poetry(self) -> None:
        r = self.__command(self.venv_in_pj_cmd)
        if r.returncode != 0:
            raise TaskError(f"Failed to '{self.venv_in_pj_cmd}'")

    def __add_dependency(self) -> None:
        ADD_CMD = "poetry add {}"
        packages = ["single-source"]
        command_list = [ADD_CMD.format(package) for package in packages]
        for cmd in command_list:
            r = self.__command(cmd)
            if r.returncode != 0:
                raise TaskError(f"Failed to add dependency: {cmd}")

    def __add_dev_dependency(self) -> None:
        ADD_CMD = "poetry add -D {}"
        dev_packages = [
            self.setting.formatter,
            self.setting.linter,
            self.setting.type_checker,
            self.setting.import_sorter,
            self.setting.unittest_framework,
            self.setting.test_framework,
        ]
        command_list = [ADD_CMD.format(package) for package in dev_packages]
        for plugin in self.setting.plugin:
            command_list.append(ADD_CMD.format(plugin))
        if self.setting.precommit:
            command_list.append(ADD_CMD.format("pre-commit"))
        for cmd in command_list:
            r = self.__command(cmd)
            if r.returncode != 0:
                raise TaskError(f"Failed to add dev dependency: {cmd}")

    def __overwrite_files(self) -> None:
        # __init__.py
        init_path = "package/init.py"
        init = get_my_resource(init_path)
        with self.path.package_init.open(mode="w") as f:
            f.write(init)
        # test_{package}.py
        test_path = "package/test_package.py"
        test = get_my_resource(test_path).replace("PACKAGE_NAME", self.setting.package_name)
        with self.path.package_test.open(mode="w") as f:
            f.write(test)



"""
This file provides environment information and prerequisite checker.
"""
import platform
import shutil
import sys
from dataclasses import dataclass
from enum import Enum, auto
from subprocess import PIPE, run
from typing import Optional

from .const import POETRY_INSTALL_GUIDE
from .cui import ask_yN


@dataclass
class Version(object):
    version: str
    separator: str = "."
    major: str = ""
    minor: str = ""
    patch: str = ""
    short: str = ""

    def __post_init__(self) -> None:
        v = self.version.split(self.separator)
        if len(v) == 3:
            self.major, self.minor, self.patch = v
        elif len(v) == 2:
            self.major, self.minor = v
        else:
            raise ValueError("The version does not contain minor version.")
        self.short = f"{self.major}{self.separator}{self.minor}"

    def __str__(self) -> str:
        return self.version


class Platform(Enum):
    WINDOWS = auto()
    UNIX = auto()
    OTHERS = auto()

    def __str__(self) -> str:
        return self.name


class Environment(object):
    def __init__(self) -> None:
        self.os: Platform = self.get_os()
        self.python: Version = self.get_python()
        self.poetry: Optional[Version] = self.get_poetry()
        self.warn_if_windows()

    def get_os(self) -> Platform:
        pf = platform.system()
        if pf == "Windows":
            return Platform.WINDOWS
        elif pf in ["Darwin", "Linux"]:
            return Platform.UNIX
        else:
            return Platform.OTHERS

    def get_python(self) -> Version:
        return Version(platform.python_version())

    def quit_poetry_not_found(self) -> None:
        print("Error: Poetry was not found.")
        print("Please install poetry first.")
        print(f"Reference: {POETRY_INSTALL_GUIDE}")
        sys.exit(1)

    def get_poetry(self) -> Optional[Version]:
        POETRY = "poetry"
        POETRY_VER = "poetry --version"
        poetry_path = shutil.which(POETRY)
        if poetry_path is None:
            self.quit_poetry_not_found()
        r = run(POETRY_VER, shell=True, stdout=PIPE, stderr=PIPE, text=True)
        if r.returncode != 0:
            self.quit_poetry_not_found()
        return Version(r.stdout.split()[2])  # "Poetry version 1.1.7"

    def warn_if_windows(self) -> None:
        if self.os is not Platform.UNIX:
            print("This operating system is not supported but you can try.")
            if not ask_yN("Do you want to proceed?"):
                print("Canceld.")
                sys.exit(0)



from ..const import INDENT, Emoji
from ..file_path import PypjFilePath
from ..resource import get_my_resource
from ..setting import PypjSetting
from .task import Task


class PreCommit(Task):
    def __init__(self, setting: PypjSetting, path: PypjFilePath) -> None:
        name = "Pre-commit"
        super().__init__(name, setting, path)

    def execute(self) -> None:
        print("Task: Configure pre-commit ")
        self.done_check()
        precommit = get_my_resource(".pre-commit-config.yaml")
        with self.path.precommit.open(mode="w") as f:
            f.write(precommit)
        print(f"{INDENT}Create : .pre-commit-config.yaml {Emoji.OK}")
        self.done()



from ..const import INDENT, Emoji
from ..file_path import PypjFilePath
from ..resource import get_my_resource
from ..setting import PypjSetting
from .task import Task


class Readme(Task):
    def __init__(self, setting: PypjSetting, path: PypjFilePath) -> None:
        name = "Readme"
        super().__init__(name, setting, path)

    def execute(self) -> None:
        print("Task: Create README.md")
        self.done_check()
        # delete README.rst
        self.path.readme_rst.unlink()
        # write README.md
        readme = get_my_resource("README.md")
        with self.path.readme_md.open(mode="w") as f:
            f.write(readme)
        print(f"{INDENT}Create : README.md {Emoji.OK}")
        self.done()



import pkgutil
from pathlib import Path

from .exception import Emsg, PypjError


def get_resource(package_name: str, resource_file: Path) -> str:
    try:
        resource = pkgutil.get_data(package_name, str(resource_file))
        if resource is None:
            raise PypjError(Emsg.RESOURCE_NOT_FOUND)
        return resource.decode("utf-8")
    except FileNotFoundError:
        raise PypjError(Emsg.RESOURCE_NOT_FOUND)


def get_my_resource(resource_file: str) -> str:
    return get_resource(__package__, Path("resources").joinpath(resource_file))



from __future__ import annotations

import re
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from .cui import ask_with_default_num, ask_Yn, ask_yN
from .environment import Version
from .exception import Emsg, PypjError
from .type_def import Formatter, ImportSorter, Linter, Plugin, TestFramework, TypeChecker


class PackageName(str):
    @staticmethod
    def is_valid(name: str) -> bool:
        regex = r"^[a-zA-Z][0-9a-zA-Z\-_]*"
        return bool(re.fullmatch(regex, name))

    @staticmethod
    def strict_format(package_name: str) -> str:
        return package_name.replace("-", "_").lower()

    def __new__(cls, name: str) -> PackageName:
        if not PackageName.is_valid(name):
            raise PypjError(Emsg.INVALID_PACKAGE_NAME)
        self = super().__new__(cls, PackageName.strict_format(name))
        return self


@dataclass
class PypjSetting(object):
    python_version: Version
    package_name: PackageName
    max_line_length: int = 119
    # use or not
    use_src: bool = False
    venv_in_pj: bool = True
    guthub_actions: bool = True
    vscode: bool = True
    makefile: bool = True
    precommit: bool = True
    # software
    formatter: Formatter = Formatter.BLACK
    linter: Linter = Linter.PFLAKE8
    type_checker: TypeChecker = TypeChecker.MYPY
    import_sorter: ImportSorter = ImportSorter.ISORT
    unittest_framework: TestFramework = TestFramework.PYTEST
    test_framework: TestFramework = TestFramework.TOX
    plugin: List[Plugin] = field(
        default_factory=lambda: [
            Plugin.FLAKE8_BUGBEAR,
            Plugin.PYTEST_COV,
            Plugin.PYTEST_MOCK,
            Plugin.TOX_GH_ACTIONS,
        ]
    )

    def __post_init__(self) -> None:
        self.validate_package_dir()
        self.src_dir: str = self.package_name  # as default

    def customize(self) -> None:
        self.max_line_length = ask_with_default_num("Max line length (119): ", 119)
        self.use_src = ask_yN("Use src directory")
        self.src_dir = "src" if self.use_src else self.package_name
        self.venv_in_pj = ask_Yn("Keep venv in project")
        self.guthub_actions = ask_Yn("Use github workflows")
        self.vscode = ask_Yn("Use vscode settings")
        self.precommit = ask_Yn("Use pre-commit")
        self.makefile = ask_Yn("Use command alias as Makefile")

    def validate_package_dir(self) -> None:
        if Path().cwd().joinpath(str(self.package_name)).exists():
            print(f"ERROR: Directory {self.package_name} already exists.")
            sys.exit(1)


FLAKE8_SETTING = """
[tool.flake8]
max-line-length = {line_length}
max-complexity = 10
select = "C,E,F,W,B"
ignore = "E203"
"""

BLACK_SETTING = """
[tool.black]
line-length = {line_length}
exclude = '''
(
    migrations
    | .mypy_cache
    | .pytest_cache
    | .tox
    | .venv
    | dist
)
'''
"""

MYPY_SETTING = """
[tool.mypy]
# common
python_version = {python_version}
show_column_numbers = true
show_error_context = true
ignore_missing_imports = true
check_untyped_defs = true
disallow_untyped_defs = true
# warning
warn_return_any = true
warn_unused_configs = true
warn_redundant_casts = true
"""

ISORT_SETTING = """
[tool.isort]
profile = "{formatter}"
line_length = {line_length}
"""

TOX_SETTING = """
[tool.tox]
legacy_tox_ini = \"\"\"
[tox]
envlist = PY3X_VERSION, flake8, black, mypy, isort
skipsdist = true
isolated_build = true
skip_missing_interpreters = true
[testenv]
whitelist_externals = poetry
require_locked_deps = true
install_dev_deps = true
commands =
    poetry install -vv --no-root
    pytest ./tests -v --cov=PACKAGE_DIR --cov-branch --durations=0
[testenv:flake8]
commands = poetry run pflake8 --statistics ./PACKAGE_DIR ./tests
[testenv:black]
commands = poetry run black ./PACKAGE_DIR ./tests
[testenv:mypy]
commands = poetry run mypy ./PACKAGE_DIR ./tests
[testenv:isort]
commands = poetry run isort ./PACKAGE_DIR ./tests
\"\"\"
"""



#!/usr/bin/env python3
import sys
from pathlib import Path
from traceback import format_exc

from pypj.cui import ask_no_empty, ask_yN, confirm_proceed

from .args import args
from .const import ASCII_ART, GITHUB_URL
from .environment import Environment
from .exception import PypjError
from .file_path import PypjFilePath
from .setting import PackageName, PypjSetting
from .task import Poetry, TaskManager


def ask_package_name() -> PackageName:
    r = ask_no_empty("Package name: ")
    return PackageName(r) if PackageName.is_valid(r) else ask_package_name()


def process() -> None:
    # preprocess
    env = Environment()
    print(ASCII_ART.format(python=env.python, poetry=env.poetry))
    # configure
    setting = PypjSetting(python_version=env.python, package_name=ask_package_name())
    if ask_yN("Do you want to customize settings?"):
        setting.customize()
    pypj_file_path = PypjFilePath(Path().cwd(), setting)
    # define tasks
    tm = TaskManager(setting, pypj_file_path)
    # execute
    confirm_proceed()
    Poetry(setting, pypj_file_path).execute()
    tm.execute()


def main() -> None:
    try:
        args()
        process()
    except PypjError as e:
        print(f"Error: {e}")
        sys.exit(1)
    except KeyboardInterrupt:
        print()
        print("Canceled.")
        sys.exit(1)
    except Exception:
        print(format_exc())
        print()
        print("If you are behind a proxy, try to set following environmental variables.")
        print("http_proxy, https_proxy, HTTP_PROXY, HTTPS_PROXY")
        print(f"Else, please report the issue to {GITHUB_URL}")
        sys.exit(1)



from .githubactions import GithubActions  # noqa
from .makefile import Makefile  # noqa
from .poetry import Poetry  # noqa
from .precommit import PreCommit  # noqa
from .pyproject import Pyproject  # noqa
from .readme import Readme  # noqa
from .task import Task, TaskStatus  # noqa
from .task_manager import TaskManager, Tasks  # noqa
from .vscode import Vscode  # noqa



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypj', 'pypj.resources.package', 'pypj.task']

package_data = \
{'': ['*'], 'pypj': ['resources/*']}

install_requires = \
['single-source>=0.2,<0.4']

entry_points = \
{'console_scripts': ['pypj = pypj.main:main']}

setup_kwargs = {
    'name': 'pypj',
    'version': '1.1.3',
    'description': 'Python project initializer',
    'long_description': '![pypj Logo](https://raw.githubusercontent.com/edge-minato/pypj/main/doc/img/logo.png)\n\n[![pypi version](https://img.shields.io/pypi/v/pypj.svg?style=flat)](https://pypi.org/pypi/pypj/)\n[![python versions](https://img.shields.io/pypi/pyversions/pypj.svg?style=flat)](https://pypi.org/pypi/pypj/)\n[![license](https://img.shields.io/pypi/l/pypj.svg?style=flat)](https://github.com/edge-minato/pypj/blob/master/LICENSE)\n[![Unittest](https://github.com/edge-minato/pypj/actions/workflows/unittest.yml/badge.svg)](https://github.com/edge-minato/pypj/actions/workflows/unittest.yml)\n[![codecov](https://codecov.io/gh/edge-minato/pypj/branch/main/graph/badge.svg?token=YDZAMKUNS0)](https://codecov.io/gh/edge-minato/pypj)\n[![Code style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black")\n[![Downloads](https://pepy.tech/badge/pypj)](https://pepy.tech/project/pypj)\n[![Downloads](https://pepy.tech/badge/pypj/week)](https://pepy.tech/project/pypj)\n\n\n`pypj` provides you a modern python project template. All the basic dev package installations, their configurations, and test workflows will be done, so you can focus on coding.\n\n## What will be provided\n\nWhat the _"Modern"_ means is expressed as following directory structure. Some developers prefer another tools, and you can remove or customize the tools to be prepared. Most of all configurations regarding the code style tools, like formatter and linter, are aggregated in `pyproject.toml`.\n\n```\nmy-package/\n├── .github\n│   ├── dependabot.yml       # Dependency updater\n│\xa0\xa0 └── workflows\n│\xa0\xa0     ├── publish.yml      # Tagging on GitHub triggers publishing to Pypi\n│\xa0\xa0     └── unittest.yml     # On each push and PR, execute the unittest\n├── .pre-commit-config.yaml  # Checks format and styles of each file\n├── .venv                    # venv can be outside of project directory\n├── .vscode\n│\xa0\xa0 └── settings.json        # Format, Lint, Type check and Import sort on save\n├── Makefile                 # Useful command alias\n├── README.md                # How to start with pypj\n├── my-package               # Your package, can be "src"\n├── poetry.lock\n├── pyproject.toml           # Configured settings\n└── tests\n```\n\n## Developing tools pypj provides\n\n- Package manager: [`poetry`](https://github.com/python-poetry/poetry)\n- Formatter: [`black`](https://github.com/psf/black)\n- Linter: [`pflake8`](https://github.com/csachs/pyproject-flake8) (\\*1)\n  - Plugin: [`flake8-bugbear`](https://github.com/PyCQA/flake8-bugbear)\n- Type checker: [`mypy`](https://github.com/python/mypy)\n- Import sorter: [`isort`](https://github.com/PyCQA/isort)\n- Test framework:\n  - [`pytest`](https://github.com/pytest-dev/pytest)\n    - [`pytest-cov`](https://github.com/pytest-dev/pytest-cov)\n    - [`pytest-mock`](https://github.com/pytest-dev/pytest-mock)\n  - [`tox`](https://github.com/tox-dev/tox)\n    - [`tox-gh-actions`](https://github.com/ymyzk/tox-gh-actions)\n- Git hooks manager: [`pre-commit`](https://github.com/pre-commit/pre-commit)\n\n(\\*1) `pflake8` wraps `flake8` to aggregate settings to `pyproject.toml`\n\n## Coding format pypj provides\n\n- Max line length: `119` as default\n- Type hinting: `required`\n- And some detailed configurations\n\n## Customize\n\nHere is an actual interaction to customize.\n\n```\nPackage name: my-package\nDo you want to customize settings? (y/N): y\nMax line length (119):\nUse src directory (y/N):\nKeep venv in project (Y/n):\nUse github workflows (Y/n):\nUse vscode settings (Y/n):\nUse pre-commit (Y/n):\nUse command alias as Makefile (Y/n):\nAre you sure? (Y/n): y\n```\n\n## Other features\n\n- Single filed configurations on `pyproject.toml`\n- Single sourced versioning: [`single-source`](https://github.com/rabbit72/single-source)\n- Command alias: [`make`](https://www.gnu.org/software/make/)\n- CI/CD\n  - unittest workflow\n  - publish package workflow\n  - dependency updater configuration\n\n## Requirements\n\n- `python3`\n- `poetry` [[Installation guide](https://python-poetry.org/docs/#installation)]\n\n## Installation\n\n```sh\npip install pypj\n```\n\n## Usage\n\nSee also [README.md](pypj/resources/README.md) which will be generated with `pypj` command, it shows more actual usage.\n\n```\n$ pypj\n\n┌─┐┬ ┬┌─┐┬\n├─┘└┬┘├─┘│    python : 3.8.5\n┴   ┴ ┴ └┘    poetry : 1.1.8\n\nPackage name: my-package\nDo you want to customize settings? (y/N): N\nDo you want to proceed? (y/N): y\nTask: Initialize package: my-package\n      Command: poetry new my-package ✨\n      Poetry new done 🚀\n      Command: poetry config virtualenvs.in-project true ✨\n      Command: poetry add -D black ✨\n      Command: poetry add -D pyproject-flake8 ✨\n      Command: poetry add -D mypy ✨\n      Command: poetry add -D isort ✨\n      Command: poetry add -D pytest ✨\n      Command: poetry add -D tox ✨\n      Command: poetry add -D pytest-cov ✨\n      Command: poetry add -D pytest-mock ✨\n      Command: poetry add -D tox-gh-actions ✨\n      Configure: __init__.py  ✨\n      Create : my-package ✨\nTask: Create README.md\n      Create : README.md ✨\nTask: Configure pyproject.toml settings\n      Write  : pyproject.toml ✨\nTask: Create github actions\n      Create : unittest.yml ✨\n      Create : publish.yml ✨\n      Create : dependabot.yml ✨\nTask: Configure vscode settings\n      Create : .vscode/settings.json ✨\nTask: Create makefile\n      Create : Makefile ✨\nTask: Configure pre-commit\n      Create : .pre-commit-config.yaml ✨\n\nComplete! 🚀\nLet\'s make the world better! ✨😋🐍🌎\n```\n\n## Example configurations on `pyproject.toml`\n\nWith default setting, this kind of `pyproject.toml` file will be generated.\n\n```toml\n[tool.poetry]\nname = "my-package"\nversion = "0.1.0"\ndescription = ""\nauthors = ["you <you@example.com>"]\n\n[tool.poetry.dependencies]\npython = "^3.8"\n\n[tool.poetry.dev-dependencies]\npytest = "^5.2"\nblack = "^21.8b0"\npyproject-flake8 = "^0.0.1-alpha.2"\nmypy = "^0.910"\nisort = "^5.9.3"\ntox = "^3.24.3"\npytest-cov = "^2.12.1"\npytest-mock = "^3.6.1"\ntox-gh-actions = "^2.7.0"\n\n[build-system]\nrequires = ["poetry-core>=1.0.0"]\nbuild-backend = "poetry.core.masonry.api"\n\n[tool.black]\nline-length = 119\nexclude = \'\'\'\n(\n    migrations\n    | .mypy_cache\n    | .pytest_cache\n    | .tox\n    | .venv\n    | dist\n)\n\'\'\'\n\n[tool.flake8]\nmax-line-length = 119\nmax-complexity = 10\nselect = "C,E,F,W,B"\nignore = "E203"\n\n[tool.mypy]\n# common\npython_version = 3.8\nshow_column_numbers = true\nshow_error_context = true\nignore_missing_imports = true\ncheck_untyped_defs = true\ndisallow_untyped_defs = true\n# warning\nwarn_return_any = true\nwarn_unused_configs = true\nwarn_redundant_casts = true\n\n[tool.isort]\nprofile = "black"\nline_length = 119\n\n[tool.tox]\nlegacy_tox_ini = """\n[tox]\nenvlist = py38, flake8, black, mypy, isort\nskipsdist = true\nisolated_build = true\nskip_missing_interpreters = true\n[testenv]\nwhitelist_externals = poetry\nrequire_locked_deps = true\ninstall_dev_deps = true\ncommands =\n    poetry install -vv --no-root\n    pytest ./tests -v --cov=pypj --cov-branch --durations=0\n[testenv:flake8]\ncommands = poetry run pflake8 ./my-package\n[testenv:black]\ncommands = poetry run black ./my-package\n[testenv:mypy]\ncommands = poetry run mypy ./my-package\n[testenv:isort]\ncommands = poetry run isort ./my-package\n"""\n```\n\n## Alias as Makefile\n\n```Makefile\n.PHONY: install update clean build run debug test style\nPACKAGE := $(shell grep name pyproject.toml -m1 | awk -F" " \'{print $$3}\')\nVERSION := $(shell grep version pyproject.toml -m1 | awk -F" " \'{print $$3}\')\n\ninstall:\n        poetry install\n        poetry run pre-commit install\n\nupdate:\n        poetry update\n        poetry run pre-commit autoupdate\n\nclean:\n        rm -rf dist\n\nbuild: clean\n        poetry build\n\nrun:\n        poetry run ${PACKAGE} # Just in case the package provides a command\n\ndebug:\n        poetry run pytest ./tests -s -v --cov=pypj --cov-branch --durations=0\n\ntest:\n        poetry run tox\n\nstyle:\n        poetry run tox -e black,flake8,mypy,isort\n```\n\n## Supported python versions\n\n- Supported: `3.7`, `3.8`, `3.9`\n- Is going to be supported: `3.10`\n- Not supported: `3.6` or less\n\n**NOTE**: According to [Status of Python branches](https://devguide.python.org/#status-of-python-branches), the EoL of Python 3.6 is `2021-12-23`.\n\n\n## FAQ\n\n* Is there any restrictions regarding the package naming?\n  * -> A name of python package is defined at [PEP-008 #Package and Module Names](https://www.python.org/dev/peps/pep-0008/#package-and-module-names) and it can be expressed as regex: `/^[a-zA-Z][0-9a-zA-Z\\-_]*/`. `pypj` follows this rule.\n',
    'author': 'edge-minato',
    'author_email': 'edge.minato@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/edge-minato/pypj',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path

from .setting import PypjSetting


class PypjFilePath(object):
    def __init__(self, pwd: Path, setting: PypjSetting) -> None:
        root = pwd.resolve()
        package_name = setting.package_name
        self.root: Path = root  # pwd as absolute path
        self.package_dir = PD = root.joinpath(package_name)  # PD as alias
        self.package_src_dir: Path = PD.joinpath(setting.src_dir)
        self.package_init: Path = self.package_src_dir.joinpath("__init__.py")
        self.package_test: Path = PD.joinpath("tests").joinpath(f"test_{package_name}.py")
        self.pyproject: Path = PD.joinpath("pyproject.toml")
        self.vscode_dir: Path = PD.joinpath(".vscode")
        self.vscode_settings_json: Path = self.vscode_dir.joinpath("settings.json")
        self.makefile: Path = PD.joinpath("Makefile")
        self.readme_md: Path = PD.joinpath("README.md")
        self.readme_rst: Path = PD.joinpath("README.rst")
        self.github_dir: Path = PD.joinpath(".github")
        self.github_workflow_dir: Path = self.github_dir.joinpath("workflows")
        self.wf_unittest: Path = self.github_workflow_dir.joinpath("unittest.yml")
        self.wf_publish: Path = self.github_workflow_dir.joinpath("publish.yml")
        self.dependabot: Path = self.github_dir.joinpath("dependabot.yml")
        self.precommit: Path = PD.joinpath(".pre-commit-config.yaml")

    def print(self) -> None:
        print(vars(self))
